#!/usr/bin/env python3
"""
Signal analysis module for phys2cvr.

Attributes
----------
LGR :
    Logger
"""

import logging

import numpy as np
import matplotlib.pyplot as plt
import scipy.interpolate as spint
import scipy.stats as sct
from scipy.signal import butter, filtfilt

from phys2cvr.io import SET_DPI, FIGSIZE


LGR = logging.getLogger(__name__)
LGR.setLevel(logging.INFO)


def spc(ts):
    """
    Compute signal percentage change of ts.

    Those ts that have mean 0 are divided by 1 instead.

    Parameters
    ----------
    ts : numpy.ndarray
        A timeseries or a set of timeseries - it is assumed that the last dimension is time.

    Returns
    -------
    numpy.ndarray
        The SPC version of ts.
    """
    m = ts.mean(axis=-1)[..., np.newaxis]
    md = m
    md[md == 0] = 1
    ts = (ts - m) / md
    ts[np.isnan(ts)] = 0

    return ts


def create_hrf(freq=40):
    """
    Create a canonical haemodynamic response function sampled at the given frequency.

    Parameters
    ----------
    freq : float
        Sampling frequency of the haemodynamic response function.

    Returns
    -------
    hrf : np.ndarray
        Haemodynamic response function.
    """
    # Create HRF
    RT = 1/freq
    fMRI_T = 16
    p = [6, 16, 1, 1, 6, 0, 32]

    # Modelled hemodynamic response function - {mixture of Gammas}
    dt = RT / fMRI_T
    u = np.arange(0, p[6]/dt+1, 1) - p[5]/dt
    a1 = p[0] / p[2]
    b1 = 1 / p[3]
    a2 = p[1] / p[3]
    b2 = 1 / p[3]
    hrf = (sct.gamma.pdf(u*dt, a1, scale=b1) - sct.gamma.pdf(u*dt, a2, scale=b2)/p[4])/dt
    time_axis = np.arange(0, int(p[6]/RT+1), 1) * fMRI_T
    hrf = hrf[time_axis]
    min_hrf = 1e-9*min(hrf[hrf > 10*np.finfo(float).eps])

    if min_hrf < 10*np.finfo(float).eps:
        min_hrf = 10*np.finfo(float).eps

    hrf[hrf == 0] = min_hrf
    hrf = hrf/max(hrf)

    return hrf


def filter_signal(data, tr, lowcut='', highcut=''):
    """
    Create a bandpass filter given a lowcut and a highcut, then filter data.

    Parameters
    ----------
    data : np.ndarray
        Data to filter (along last axis)
    tr : float
        TR of functional files
    lowcut : float
        Lower frequency in the bandpass
    highcut : float
        Higher frequency in the bandpass

    Returns
    -------
    filt_data : np.ndarray
        Input `data`, but filtered.
    """
    if not lowcut:
        lowcut = 0.02
    if not highcut:
        highcut = 0.04
    nyq = (1 / tr) / 2
    low = lowcut / nyq
    high = highcut / nyq
    a, b = butter(1, [low, high], btype='band')
    filt_data = filtfilt(a, b, data, axis=-1)
    return filt_data


def convolve_petco2(co2, pidx, freq, outname):
    """
    Convolve the CO2 trace into the PetCO2 trace.

    Parameters
    ----------
    co2 : np.ndarray
        CO2 (or physiological) regressor
    pidx : np.ndarray
        index of peaks
    freq : str, int, or float
        sample frequency of the CO2 regressor
    outname : str
        prefix of the exported file

    Returns
    -------
    co2_conv : np.ndarray
        Convolved CO2 trace
    """
    # Extract PETco2
    hrf = create_hrf(freq)
    co2_lenght = len(co2)
    nx = np.linspace(0, co2_lenght, co2_lenght)
    f = spint.interp1d(pidx, co2[pidx], fill_value='extrapolate')
    petco2 = f(nx)

    # Plot PETco2
    plt.figure(figsize=FIGSIZE, dpi=SET_DPI)
    plt.title('CO2 and PetCO2')
    plt.plot(co2, '-', petco2, '-')
    plt.savefig(f'{outname}_petco2.png', dpi=SET_DPI)
    plt.close()

    # Demean and export
    petco2 = petco2 - petco2.mean()
    np.savetxt(f'{outname}_petco2.1D', petco2, fmt='%.18f')

    # Convolve, and then rescale to have same amplitude (?)
    co2_conv = np.convolve(petco2, hrf)
    co2_conv = np.interp(co2_conv, (co2_conv.min(), co2_conv.max()),
                         (petco2.min(), petco2.max()))

    plt.figure(figsize=FIGSIZE, dpi=SET_DPI)
    plt.title('PetCO2 and convolved regressor (PetCO2hrf)')
    plt.plot(co2_conv, '-', petco2, '-')
    plt.savefig(f'{outname}_petco2hrf.png', dpi=SET_DPI)
    plt.close()

    np.savetxt(f'{outname}_petco2hrf.1D', co2_conv, fmt='%.18f')

    return co2_conv


def resample_signal(ts, freq1, freq2):
    """
    Upsample or downsample a given timeseries.

    This program brings ts at freq1 to a new timeseries at freq2

    Parameters
    ----------
    ts : numpy.ndarray
        The timeseries to resample
    freq1 : float
        The frequency of the timeseries to resample
    freq2 : float
        The new desired frequency

    Returns
    -------
    numpy.ndarray
        The resampled timeseries
    """
    # Upsample functional signal
    len_tp = ts.shape[-1]
    len_s = (len_tp - 1) * 1/freq1
    regr_t = np.linspace(0, len_s, int(len_s*freq2)+1)
    time_t = np.linspace(0, len_s, len_tp)
    f = spint.interp1d(time_t, ts, fill_value='extrapolate')

    return f(regr_t)


"""
Copyright 2021, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""
Main file for phys2cvr.

Attributes
----------
LGR :
    Logger
"""

import datetime
import logging
import os
import sys
from copy import deepcopy

import numpy as np
from peakdet.io import load_physio

from phys2cvr import io, signal, stats, _version
from phys2cvr.cli.run import _get_parser, _check_opt_conf
from phys2cvr.io import EXT_NIFTI, EXT_1D


LGR = logging.getLogger(__name__)
LGR.setLevel(logging.INFO)


def save_bash_call(fname, outdir):
    """
    Save the bash call into file `p2d_call.sh`.

    Parameters
    ----------
    outdir : str or path, optional
        output directory
    """
    arg_str = ' '.join(sys.argv[1:])
    call_str = f'phys2cvr {arg_str}'
    if outdir:
        outdir = os.path.abspath(outdir)
    else:
        outdir = os.path.join(os.path.split(fname)[0], 'phys2cvr')
    log_path = os.path.join(outdir, 'logs')
    os.makedirs(log_path, exist_ok=True)
    isotime = datetime.datetime.now().strftime('%Y-%m-%dT%H%M%S')
    fname, _ = io.check_ext('.nii.gz', os.path.basename(fname), remove=True)
    f = open(os.path.join(log_path, f'p2c_call_{fname}_{isotime}.sh'), 'a')
    f.write(f'#!bin/bash \n{call_str}')
    f.close()


def phys2cvr(fname_func, fname_co2=None, fname_pidx=None, fname_roi=None, fname_mask=None,
             outdir=None, freq=None, tr=None, trial_len=None, n_trials=None,
             highcut=None, lowcut=None, apply_filter=False,
             run_regression=False, lagged_regression=True, r2model='full', lag_max=None,
             lag_step=None, legacy=False, l_degree=0, denoise_matrix=[], scale_factor=None,
             lag_map=None, regr_dir=None, run_conv=True, quiet=False, debug=False):
    """
    Run main workflow of phys2cvr.

    Parameters
    ----------
    fname_func : str or path
        Filename of the functional input (nifti or txt)
    fname_co2 : str or path, optional
        Filename of the CO2 (physiological regressor) timeseries.
        Can be either peakdet's output or a txt file.
        If not declared, phys2cvr will consider the average temporal value
        from the input.
        Default: empty
    fname_pidx : str or path, optional
        Filename of the CO2 (physiological regressor) timeseries' PEAKS.
        Required if CO2 file is a txt AND the convolution step is not skipped.
        If not declared AND the convolution step is not skipped, raises an exception.
        Default: empty
    fname_roi : str or path, optional
        Filename of the roi in a nifti volume.
        If declared, phys2cvr will use these voxels .
        If not, phys2cvr will use a mask, either the declared one or estimated
        from the functional input.
        Ignored if input is a txt file.
        Default: empty
    fname_mask : str or path, optional
        Filename of the mask in a nifti volume.
        If declared, phys2cvr will run only on these voxels.
        If not, phys2cvr will estimate a mask from the functional input.
        Ignored if input is a txt file.
        Default: empty
    outdir : str or path, optional
        Output directory
        Default: the directory where `fname_func` is.
    freq : str, int, or float, optional
        Sample frequency of the CO2 regressor. Required if CO2 input is TXT file.
        If declared with peakdet file, it will overwrite the file frequency.
    tr : str, int, or float, optional
        TR of the timeseries. Required if input is TXT file.
        If declared with nifti file, it will overwrite the file TR.
    trial_len : str or int, optional
        Length of each single trial for tasks that have more than one
        (E.g. BreathHold, CO2 challenges, ...)
        Used to improve cross correlation estimation.
        Default: None
    n_trials : str or int, optional
        Number of trials in the task.
        Default: None
    highcut : str, int, or float, optional
        High frequency limit for filter.
        Required if applying a filter.
        Default: empty
    lowcut : str, int, or float, optional
        Low frequency limit for filter.
        Required if applying a filter.
        Default: empty
    apply_filter : bool, optional
        Apply a Butterworth filter to the functional input.
        Default: False
    run_regression : bool, optional
        Also run the regression step within phys2cvr.
        By default, phys2cvr will only estimate the regressor(s) of interest.
        Default: False
    lagged_regression : bool, optional
        Estimates regressors to run a lagged regression approach.
        If `run_regression` is True, also run the lagged regression.
        Can be turned off.
        Default: True
    r2model : 'full', 'partial' or 'intercept', optional
        Submit the model of the R^2 the regression should return (hence, which
        R^2 model the lag regression should be based on).
        Possible options are 'full', 'partial', 'intercept'.
        See `stats.regression` help to understand them.
        Default: 'full'
    lag_max : int or float, optional
        Limits (both positive and negative) of the temporal area to explore,
        expressed in seconds (e.g. ±9 seconds). Caution: this is not a pythonic
        range, but a real range, i.e. the upper limit is included (e.g. [-9, +9]).
        Default: None
    lag_step : int or float, optional
        Step of the lag to take into account in seconds.
        Default: None
    legacy : bool, optional
        If True, use pythonic ranges when creating the regressors, i.e. exclude
        the upper range (e.g. [-9, +9) ).
        Default: False
    l_degree : int, optional
        Only used if performing the regression step.
        Highest order of the Legendre polynomial to add to the denoising matrix.
        phys2cvr will add all polynomials up to the specified order
        (e.g. if user specifies 3, orders 0, 1, 2, and 3 will be added).
        Default is 0, which will model only the mean of the timeseries.
    denoise_matrix : list of str(s) or path(s), optional
        Add one or multiple denoising matrices to the regression model.
        Ignored if not performing the regression step.
    scale_factor : str, int, or float, optional
        A scale factor to apply to the CVR map before exporting it.
        Useful when using inputs recorded/stored in Volts that have a meaningful
        unit of measurement otherwise, e.g. (CO2 traces in mmHg).
        V->mmHg: CO2[mmHg] = (Patm-Pvap)[mmHg] * 10*CO2[V]/100[V]
        Default: None
    lag_map : str or path, optional
        Filename of a lag map to get lags from.
        Ignored if not running a lagged-GLM regression.
        Default: None
    regr_dir : str, optional
        Directory containing pre-generated lagged regressors, useful
        to (re-)run a GLM analysis.
        Default: None
    run_conv : bool, optional
        Run the convolution of the physiological trace.
        Can be turned off
        Default: True
    quiet : bool, optional
        Return to screen only warnings and errors.
        Default: False
    debug : bool, optional
        Return to screen more output.
        Default: False

    Raises
    ------
    Exception
        - If functional timeseries is lacking TR and the latter was not specified.
        - If functional nifti file is not at least 4D.
        - If mask was specified but it has different dimensions than the
            functional nifti file.
        - If a file type is not supported yet.
        - If physiological file is a txt file and no peaks were provided.
        - If physiological file is lacking frequency and the latter was not specified.
        - If a lag map was specified but it has different dimensions than the
            functional nifti file.
        - If a lag map was specified, lag_step was not, and the lag map seems
            to have different lag_steps inside.
    """
    # If lagged regression is selected, make sure run_regression is true.
    if lagged_regression:
        run_regression = True
    # Add logger and suff
    if outdir:
        outdir = os.path.abspath(outdir)
    else:
        outdir = os.path.join(os.path.split(fname_func)[0], 'phys2cvr')
    outdir = os.path.abspath(outdir)
    petco2log_path = os.path.join(outdir, 'logs')
    os.makedirs(petco2log_path, exist_ok=True)

    # Create logfile name
    basename = 'phys2cvr_'
    extension = 'tsv'
    isotime = datetime.datetime.now().strftime('%Y-%m-%dT%H%M%S')
    logname = os.path.join(petco2log_path, f'{basename}{isotime}.{extension}')

    # Set logging format
    log_formatter = logging.Formatter(
        '%(asctime)s\t%(name)-12s\t%(levelname)-8s\t%(message)s',
        datefmt='%Y-%m-%dT%H:%M:%S')

    # Set up logging file and open it for writing
    log_handler = logging.FileHandler(logname)
    log_handler.setFormatter(log_formatter)
    sh = logging.StreamHandler()

    if quiet:
        logging.basicConfig(level=logging.WARNING,
                            handlers=[log_handler, sh], format='%(levelname)-10s %(message)s')
    elif debug:
        logging.basicConfig(level=logging.DEBUG,
                            handlers=[log_handler, sh], format='%(levelname)-10s %(message)s')
    else:
        logging.basicConfig(level=logging.INFO,
                            handlers=[log_handler, sh], format='%(levelname)-10s %(message)s')

    version_number = _version.get_versions()['version']
    LGR.info(f'Currently running phys2cvr version {version_number}')
    LGR.info(f'Input file is {fname_func}')

    # Check func type and read it
    func_is_1d = io.check_ext(EXT_1D, fname_func)
    func_is_nifti = io.check_ext(EXT_NIFTI, fname_func)

    # Check that all input values have right type
    tr = io.if_declared_force_type(tr, 'float', 'tr')
    freq = io.if_declared_force_type(freq, 'float', 'freq')
    trial_len = io.if_declared_force_type(trial_len, 'int', 'trial_len')
    n_trials = io.if_declared_force_type(n_trials, 'int', 'n_trials')
    highcut = io.if_declared_force_type(highcut, 'float', 'highcut')
    lowcut = io.if_declared_force_type(lowcut, 'float', 'lowcut')
    lag_max = io.if_declared_force_type(lag_max, 'float', 'lag_max')
    lag_step = io.if_declared_force_type(lag_step, 'float', 'lag_step')
    l_degree = io.if_declared_force_type(l_degree, 'int', 'l_degree')
    if l_degree < 0:
        raise ValueError('The specified order of the Legendre polynomials must be >= 0.')
    scale_factor = io.if_declared_force_type(scale_factor, 'float', 'scale_factor')
    if r2model not in stats.R2MODEL:
        raise ValueError(f'R^2 model {r2model} not supported. Supported models '
                         f'are {stats.R2MODEL}')

    if func_is_1d:
        if tr:
            func_avg = np.genfromtxt(fname_func)
            LGR.info(f'Loading {fname_func}')
            if apply_filter:
                LGR.info('Applying butterworth filter to {fname_func}')
                func_avg = signal.filter_signal(func_avg, tr, lowcut, highcut)
        else:
            raise NameError('Provided functional signal, but no TR specified! '
                            'Rerun specifying the TR')
    elif func_is_nifti:
        func, dmask, img = io.load_nifti_get_mask(fname_func)
        if len(func.shape) < 4:
            raise ValueError(f'Provided functional file {fname_func} is not a 4D file!')
        # Read TR or declare its overwriting
        if tr:
            LGR.warning(f'Forcing TR to be {tr} seconds')
        else:
            tr = img.header['pixdim'][4]

        # Read mask (and mask func) if provided
        if fname_mask:
            _, mask, _ = io.load_nifti_get_mask(fname_mask, is_mask=True)
            if func.shape[:3] != mask.shape:
                raise ValueError(f'{fname_mask} and {fname_func} have different sizes!')
            mask = mask * dmask
            LGR.info(f'Masking {os.path.basename(fname_func)} using {os.path.basename(fname_mask)}')
            func = func * mask[..., np.newaxis]
            roiref = os.path.basename(fname_mask)
        else:
            mask = dmask
            LGR.warning(f'No mask specified, using any voxel different from 0 in '
                        f'{os.path.basename(fname_func)}')
            roiref = os.path.basename(fname_func)

        # Read roi if provided
        if fname_roi:
            _, roi, _ = io.load_nifti_get_mask(fname_roi, is_mask=True)
            if func.shape[:3] != roi.shape:
                raise ValueError(f'{fname_roi} and {fname_func} have different sizes!')
            roi = roi * mask
            roiref = os.path.basename(fname_roi)
        else:
            roi = mask
            LGR.warning(f'No ROI specified, using any voxel different from 0 in '
                        f'{roiref}')

        if apply_filter:
            LGR.info(f'Obtaining filtered average signal in {roiref}')
            func_filt = signal.filter_signal(func, tr, lowcut, highcut)
            func_avg = func_filt[roi].mean(axis=0)
        else:
            LGR.info(f'Obtaining average signal in {roiref}')
            func_avg = func[roi].mean(axis=0)

    else:
        raise NotImplementedError(f'{fname_func} file type is not supported yet, or '
                                  'the extension was not specified.')

    if fname_co2 is None:
        LGR.info(f'Computing "CVR" (approximation) maps using {fname_func} only')
        if func_is_1d:
            LGR.warning('Using an average signal only, solution might be unoptimal.')

            if apply_filter is None:
                LGR.warning('No filter applied to the input average! You know '
                            'what you are doing, right?')

        if func_is_1d:
            petco2hrf = signal.spc(func_avg)
        elif func_is_nifti:
            # Get the average of the SPC rather than the SPC of the average
            if apply_filter:
                func_filt = signal.filter_signal(func, tr, lowcut, highcut)
                petco2hrf = signal.spc(func_filt[roi]).mean(axis=0)
            else:
                petco2hrf = signal.spc(func[roi]).mean(axis=0)

        # Reassign fname_co2 to fname_func for later use - calling splitext twice cause .gz
        basename_co2 = os.path.splitext(os.path.splitext(f'avg_{os.path.basename(fname_func)}')[0])[0]
        outname = os.path.join(outdir, basename_co2)

        # If freq was declared, upsample the average GM to that.
        # Otherwise, set freq to inverse of TR.
        if freq is None:
            freq = 1/tr
            LGR.info(f'No frequency declared, using 1/tr ({freq}Hz)')
        else:
            LGR.info(f'Resampling the average fMRI timeseries at {freq}Hz')
            petco2hrf = signal.resample_signal(petco2hrf, 1/tr, freq)
    else:
        co2_is_phys = io.check_ext('.phys', fname_co2)
        co2_is_1d = io.check_ext(EXT_1D, fname_co2)

        if co2_is_1d:
            if fname_pidx:
                pidx = np.genfromtxt(fname_pidx)
                pidx = pidx.astype(int)
            elif run_conv:
                raise NameError(f'{fname_co2} file is a text file, but no '
                                'file containing its peaks was provided. '
                                ' Please provide peak file!')

            if freq is None:
                raise NameError(f'{fname_co2} file is a text file, but no '
                                'frequency was specified. Please provide peak '
                                ' file!')

            co2 = np.genfromtxt(fname_co2)
        elif co2_is_phys:
            # Read a phys file!
            phys = load_physio(fname_co2, allow_pickle=True)

            co2 = phys.data
            pidx = phys.peaks
            if freq:
                LGR.warning(f'Forcing CO2 frequency to be {freq} Hz')
            else:
                freq = phys.fs
        else:
            raise NotImplementedError(f'{fname_co2} file type is not supported yet, or '
                                      'the extension was not specified.')

        # Set output file & path - calling splitext twice cause .gz
        basename_co2 = os.path.splitext(os.path.splitext(os.path.basename(fname_co2))[0])[0]
        outname = os.path.join(outdir, basename_co2)

        # Unless user asks to skip this step, convolve the end tidal signal.
        if run_conv is False:
            petco2hrf = co2
        else:
            petco2hrf = signal.convolve_petco2(co2, pidx, freq, outname)

    # If a regressor directory is not specified, compute the regressors.
    if regr_dir is None:
        regr, regr_shifts = stats.get_regr(func_avg, petco2hrf, tr, freq, outname,
                                           lag_max, trial_len, n_trials,
                                           '.1D', lagged_regression, legacy)
    elif run_regression:
        try:
            regr = np.genfromtxt(f'{outname}_petco2hrf.1D')
        except IOError:
            LGR.warning(f'Regressor {outname}_petco2hrf.1D not found. '
                        'Estimating it.')
            regr, regr_shifts = stats.get_regr(func_avg, petco2hrf, tr, freq,
                                               outname, lag_max, trial_len,
                                               n_trials, '.1D', lagged_regression,
                                               legacy)

    # Run internal regression if required and possible!
    if func_is_nifti and run_regression:
        LGR.info('Running regression!')

        # Change dimensions in image header before export
        LGR.info('Prepare output image')
        fname_out_func, _ = io.check_ext('.nii.gz', os.path.basename(fname_func), remove=True)
        fname_out_func = os.path.join(outdir, fname_out_func)
        newdim = deepcopy(img.header['dim'])
        newdim[0], newdim[4] = 3, 1
        oimg = deepcopy(img)
        oimg.header['dim'] = newdim

        # Compute signal percentage change of functional data
        func = signal.spc(func)

        # Start computing the polynomial regressor (at least average)
        LGR.info(f'Compute Legendre polynomials up to order {l_degree}')
        mat_conf = stats.get_legendre(l_degree, regr.size)

        # Read in eventual denoising factors
        if denoise_matrix:
            denoise_matrix = io.if_declared_force_type(denoise_matrix, 'list', 'denoise_matrix')
            for matrix in denoise_matrix:
                LGR.info(f'Read confounding factor from {matrix}')
                conf = np.genfromtxt(matrix)
                mat_conf = np.hstack([mat_conf, conf])

        LGR.info('Compute simple CVR estimation (bulk shift only)')
        x1D = os.path.join(outdir, 'mat', 'mat_simple.1D')
        beta, tstat, r_square = stats.regression(func, mask, regr, mat_conf,
                                                 r2model, debug, x1D)

        LGR.info('Export bulk shift results')
        if scale_factor is None:
            LGR.warning('Remember: CVR might not be in %BOLD/mmHg!')
        else:
            beta = beta / float(scale_factor)
        # Scale beta by scale factor while exporting (useful to transform V in mmHg)
        LGR.info('Export CVR and T-stat of simple regression')
        io.export_nifti(beta, oimg, f'{fname_out_func}_cvr_simple')
        io.export_nifti(tstat, oimg, f'{fname_out_func}_tstat_simple')

        if debug:
            LGR.debug('Export R^2 volume of simple regression')
            io.export_nifti(r_square, oimg, f'{fname_out_func}_r_square_simple')

        if lagged_regression and regr_shifts is not None and ((lag_max and lag_step) or lag_map):
            if lag_max:
                LGR.info(f'Running lagged CVR estimation with max lag = {lag_max}! '
                         '(might take a while...)')
            elif lag_map is not None:
                LGR.info(f'Running lagged CVR estimation with lag map {lag_map}! '
                         '(might take a while...)')
            if legacy:
                nrep = int(lag_max * freq * 2)
            else:
                nrep = int(lag_max * freq * 2) + 1

            if regr_dir:
                outprefix = os.path.join(regr_dir, os.path.split(outname)[1])

            # If user specified a lag map, use that one to regress things
            if lag_map:
                lag, _, _ = io.load_nifti_get_mask(lag_map)
                if func.shape[:3] != lag.shape:
                    raise ValueError(f'{lag_map} and {fname_func} have different sizes!')

                # Read lag_step and lag_max from file (or try to)
                lag = lag * mask

                lag_list = np.unique(lag)

                if lag_step is None:
                    lag_step = np.unique(lag_list[1:] - lag_list[:-1])
                    if lag_step.size > 1:
                        raise ValueError(f'phys2cvr found different delta lags in {lag_map}')
                    else:
                        LGR.warning(f'phys2cvr detected a delta lag of {lag_step} seconds')
                else:
                    LGR.warning(f'Forcing delta lag to be {lag_step}')

                step = int(lag_step * freq)

                if lag_max is None:
                    lag_max = np.abs(lag_list).max()
                    LGR.warning(f'phys2cvr detected a max lag of {lag_max} seconds')
                else:
                    LGR.warning(f'Forcing max lag to be {lag_max}')

                lag_idx = (lag + lag_max) * freq / step

                lag_idx_list = np.unique[lag_idx]

                # Prepare empty matrices
                beta = np.empty_like(lag, dtype='float32')
                tstat = np.empty_like(lag, dtype='float32')

                for i in lag_idx_list:
                    LGR.info(f'Perform L-GLM for lag {lag_list[i]} ({i+1} of '
                             f'{len(lag_idx_list)}')
                    try:
                        regr = regr_shifts[:, (i*step)]
                    except NameError:
                        regr = np.genfromtxt(f'{outprefix}_{i:04g}')

                    x1D = os.path.join(outdir, 'mat', f'mat_{i:04g}.1D')
                    (beta[lag_idx == i],
                     tstat[lag_idx == i],
                     _) = stats.regression(func[lag_idx == i], [lag_idx == i],
                                           regr, mat_conf, r2model, debug,
                                           x1D)

            else:
                # Check the number of repetitions first
                if lag_step:
                    step = int(lag_step * freq)
                else:
                    step = 1
                lag_range = list(range(0, nrep, step))
                # Prepare empty matrices
                r_square_all = np.empty(list(func.shape[:3]) + [len(lag_range)],
                                        dtype='float32')
                beta_all = np.empty(list(func.shape[:3]) + [len(lag_range)],
                                    dtype='float32')
                tstat_all = np.empty(list(func.shape[:3]) + [len(lag_range)],
                                     dtype='float32')

                for n, i in enumerate(lag_range):
                    LGR.info(f'Perform L-GLM number {n+1} of {len(lag_range)}')
                    try:
                        regr = regr_shifts[:, i]
                        LGR.debug(f'Using shift {i} from matrix in memory: {regr}')
                    except NameError:
                        regr = np.genfromtxt(f'{outprefix}_{i:04g}')
                        LGR.debug(f'Reading shift {i} from file {outprefix}_{i:04g}')

                    x1D = os.path.join(outdir, 'mat', f'mat_{i:04g}.1D')
                    (beta_all[:, :, :, n],
                     tstat_all[:, :, :, n],
                     r_square_all[:, :, :, n]) = stats.regression(func, mask,
                                                                  regr,
                                                                  mat_conf,
                                                                  r2model,
                                                                  debug,
                                                                  x1D)

                if debug:
                    LGR.debug('Export all betas, tstats, and R^2 volumes.')
                    newdim_all = deepcopy(img.header['dim'])
                    newdim_all[0], newdim_all[4] = 4, int(len(lag_range))
                    oimg_all = deepcopy(img)
                    oimg_all.header['dim'] = newdim_all
                    io.export_nifti(r_square_all, oimg_all, f'{fname_out_func}_r_square_all')
                    io.export_nifti(tstat_all, oimg_all, f'{fname_out_func}_tstat_all')
                    io.export_nifti(beta_all, oimg_all, f'{fname_out_func}_beta_all')

                # Find the right lag for CVR estimation
                lag_idx = np.argmax(r_square_all, axis=-1)
                lag = (lag_idx * step) / freq - (mask * lag_max)
                # Express lag map relative to median of the roi
                lag_rel = lag - (mask * np.median(lag[roi]))

                # Run through indexes to pick the right value
                lag_idx_list = np.unique(lag_idx)
                beta = np.empty_like(lag, dtype='float32')
                tstat = np.empty_like(lag, dtype='float32')
                for i in lag_idx_list:
                    beta[lag_idx == i] = beta_all[:, :, :, i][lag_idx == i]
                    tstat[lag_idx == i] = tstat_all[:, :, :, i][lag_idx == i]

            LGR.info('Export fine shift results')
            if scale_factor is None:
                LGR.warning('Remember: CVR might not be in %BOLD/mmHg!')
            else:
                beta = beta / float(scale_factor)

            io.export_nifti(beta, oimg, f'{fname_out_func}_cvr')
            io.export_nifti(tstat, oimg, f'{fname_out_func}_tstat')
            if not lag_map:
                io.export_nifti(lag, oimg, f'{fname_out_func}_lag')
                io.export_nifti(lag_rel, oimg, f'{fname_out_func}_lag_mkrel')

    elif run_regression:
        LGR.warning('The input file is not a nifti volume. At the moment, '
                    'regression is not supported for other formats.')

    LGR.info('phys2cvr finished! Enjoy your outputs!')


def _main(argv=None):
    options = _get_parser().parse_args(argv)

    options = _check_opt_conf(options)

    save_bash_call(options.fname_func, options.outdir)

    phys2cvr(**vars(options))


if __name__ == '__main__':
    _main(sys.argv[1:])


"""
Copyright 2021, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



"""Hopefully importing everything."""

import pkgutil

from ._version import get_versions


__version__ = get_versions()['version']
del get_versions

__all__ = []
for loader, module_name, is_pkg in pkgutil.walk_packages(__path__):
    __all__.append(module_name)
    _module = loader.find_module(module_name).load_module(module_name)
    globals()[module_name] = _module


# -*- coding: utf-8 -*-
"""
Parser for phys2cvr.
"""

import argparse

from phys2cvr import __version__


def _get_parser():
    """
    Parses command line inputs for this function

    Returns
    -------
    parser.parse_args() : argparse dict

    """
    parser = argparse.ArgumentParser(description=('%(prog)s, a tool to compute '
                                                  'Cerebrovascular Reactivity '
                                                  'maps and their lags.\n'
                                                  '%(prog)s is compatible with '
                                                  'different designs and techniques '
                                                  'to estimate CVR maps.'
                                                  'It can also be used to generate '
                                                  'regressors to run the estimation '
                                                  'with other softwares.\n'
                                                  f'Version {__version__}'),
                                     add_help=False)
    required = parser.add_argument_group('Required Argument')
    required.add_argument('-i', '--input-func',
                          dest='fname_func',
                          type=str,
                          help=('Complete path (absolute or relative) and name '
                                'of the file containing fMRI signal. This file '
                                'can be a nifti file or a 1D txt file.'),
                          required=True)

    opt_out = parser.add_argument_group('Optional Argument for output')
    opt_out.add_argument('-o', '--output-directory',
                         dest='outdir',
                         type=str,
                         help=('Complete path (absolute or relative) and name '
                               'of the desired output directory. If it does not '
                               'exist, it will be created. If it is not '
                               'specified, a folder named "phys2cvr" will be '
                               'created in the folder containing the functional '
                               'file.'),
                         default=None)

    opt_func = parser.add_argument_group('Optional Arguments for fMRI timeseries')
    opt_func.add_argument('-m', '--input-mask',
                          dest='fname_mask',
                          type=str,
                          help=('Complete path (absolute or relative) and name '
                                'of the file containing a brain mask (nifti file). '
                                'Only the voxels in this mask will be considered '
                                'by phys2cvr. Use this option to '
                                'specify a GM mask or overwrite a full brain mask.\n'
                                'If the functional file is specified and this '
                                'option is not used, or the mask cannot be '
                                'loaded, the program will create a mask using '
                                'any voxel of the functional file constantly '
                                'different from zero.'),
                          default=None)
    opt_func.add_argument('-r', '--input-roi',
                          dest='fname_roi',
                          type=str,
                          help=('Complete path (absolute or relative) and name '
                                'of the nifti file containing a subset of voxels to '
                                'treat as a region of interest (ROI). '
                                'The average functional signal of the ROI will be used '
                                'to run the cross correlation with the '
                                'physiological regressor. The median lag value in '
                                'the ROI will be used to correct the final lag map.\n'
                                'If the functional file is specified and this '
                                'option is not used, or the ROI cannot be '
                                'loaded, the program will either use a specified mask '
                                '(see `--input-mask`) or create a mask using '
                                'any voxel of the functional file constantly '
                                'different from zero.'),
                          default=None)
    opt_func.add_argument('-tr', '--repetition-time',
                          dest='tr',
                          type=float,
                          help=('TR of functional data. Required '
                                'if the latter is not passed as a nifti file. '
                                'Use this option to overwrite the frequency of a '
                                'nifti file.'),
                          default=None)

    opt_phys = parser.add_argument_group('Optional Arguments for physiological '
                                         'timeseries (regressor of interest)')
    opt_phys.add_argument('-co2', '--input-co2',
                          dest='fname_co2',
                          type=str,
                          help=('Complete path (absolute or relative) and name '
                                'of the file containing CO2 signal (or equivalent '
                                'physiological trace to compute the regressor). '
                                'This file can be a 1D txt file or a .phys file '
                                'from peakdet.\n If nothing is specified, the '
                                'average timeseries of the mask will be used '
                                'as regressor.'),
                          default=None)
    opt_phys.add_argument('-pk', '--input-peaks',
                          dest='fname_pidx',
                          type=str,
                          help=('Complete path (absolute or relative) and name '
                                'of the file containing the peak of the '
                                'physiological trace. Required if the physiological '
                                'trace file is not a .phys file.\n'
                                'Use this option to overwrite the peaks specified '
                                'in the .phys file.'),
                          default=None)
    opt_phys.add_argument('-fr', '--frequency',
                          dest='freq',
                          type=float,
                          help=('Frequency of the physiological trace. Required '
                                'if the latter is not passed as a .phys file.\n'
                                'Use this option to overwrite the frequency of a '
                                '.phys file.'),
                          default=None)

    opt_xcorr = parser.add_argument_group('Optional Arguments for the cross correlation '
                                          '(bulk shift estimation step)')
    opt_xcorr.add_argument('-tlen', '--trial-length',
                           dest='trial_len',
                           type=float,
                           help=('Total duration of a single trial of the task '
                                 'in seconds (useful for block designs).\n'
                                 'Specify this with the number of trials to run '
                                 'a double cross-correlation between functional '
                                 'signal and physiological regressor to improve '
                                 'the detection of the bulk shift.'),
                           default=None)
    opt_xcorr.add_argument('-ntrial', '--trial-number',
                           dest='n_trials',
                           type=int,
                           help=('Number of trials in the task (useful for block designs).\n'
                                 'Specify this with the duration of trials to run '
                                 'a double cross-correlation between functional '
                                 'signal and physiological regressor to improve '
                                 'the detection of the bulk shift.'),
                           default=None)

    opt_filt = parser.add_argument_group('Optional Arguments for temporal filter')
    opt_filt.add_argument('-af', '--apply-filter',
                          dest='apply_filter',
                          action='store_true',
                          help=('Apply a filter to the functional data before '
                                'estimating the bulk shift. The filter will not '
                                'be applied on the data before the GLM computation. '
                                'If you want that, consider applying it before '
                                'running %(prog)s (or running it a second time).'),
                          default=False)
    opt_filt.add_argument('-hf', '--highcut-frequency',
                          dest='highcut',
                          type=float,
                          help=('Higher frequency to use in signal filtering. '
                                'The filter will be applied only to the functional '
                                'data to estimate the bulk shift. This option '
                                'is suggested when only using a functional file.'),
                          default=None)
    opt_filt.add_argument('-lf', '--lowcut-frequency',
                          dest='lowcut',
                          type=float,
                          help=('Lower frequency to use in signal filtering. '
                                'The filter will be applied only to the functional '
                                'data to estimate the bulk shift. This option '
                                'is suggested when only using a functional file.'),
                          default=None)

    opt_flow = parser.add_argument_group('Optional Arguments to modify the workflow')
    opt_flow.add_argument('-skip_reg', '--skip-regression',
                          dest='run_regression',
                          action='store_false',
                          help=('Skip running physiological regression(s) '
                                'internally. This will make %(prog)s '
                                'generate the desired physiological regressors '
                                'and quit, assuming that the regression itself '
                                'will be carried out with other software '
                                '(e.g. AFNI, FSL, ...).'),
                          default=True)
    opt_flow.add_argument('-skip_lagreg', '--skip-lagged-regression',
                          dest='lagged_regression',
                          action='store_false',
                          help=('Skip estimating the lagged regressors, '
                                'estimating only the bulk shifted one.\n'
                                'Skip running the lagged regression if the '
                                'regression step is run.'),
                          default=True)
    opt_flow.add_argument('-skip_conv', '--skip-convolution',
                          dest='run_conv',
                          action='store_false',
                          help=('Skip convolution of physiological trace. '
                                'By default %(prog)s convolves the physiological '
                                'trace with a standard HRF. Skip it when using '
                                'fMRI signal only.'),
                          default=True)

    title_opt_r2model = parser.add_argument_group('Optional Arguments to select R^2 model '
                                                  'for regresion step')

    opt_r2model = title_opt_r2model.add_mutually_exclusive_group()
    opt_r2model.add_argument('--r2full',
                             dest='r2model',
                             action='store_const',
                             const='full',
                             help=('Use full R^2 of the model, , i.e. compare versus baseline 0'),
                             default=None)
    opt_r2model.add_argument('--r2partial',
                             dest='r2model',
                             action='store_const',
                             const='partial',
                             help=('Use partial R^2 of the regressor of interest, i.e. compare '
                                   'versus any other regressor'),
                             default=None)
    opt_r2model.add_argument('--r2intercept',
                             dest='r2model',
                             action='store_const',
                             const='intercept',
                             help=('Use full R^2 of the model but the intercept, i.e. compare '
                                   'versus baseline intercept (Legendre polynomial order 0, '
                                   'a.k.a. average signal)'),
                             default=None)
    opt_r2model.add_argument('--r2adjfull',
                             dest='r2model',
                             action='store_const',
                             const='adj_full',
                             help=('Same as \'full\', but adjusted'),
                             default=None)
    opt_r2model.add_argument('--r2adjpartial',
                             dest='r2model',
                             action='store_const',
                             const='adj_partial',
                             help=('Same as \'partial\', but adjusted'),
                             default=None)
    opt_r2model.add_argument('--r2adjintercept',
                             dest='r2model',
                             action='store_const',
                             const='adj_intercept',
                             help=('Same as \'intercept\', but adjusted'),
                             default=None)

    opt_regr = parser.add_argument_group('Optional Arguments for the regression step')
    opt_regr.add_argument('-ldeg', '--legendre-degree',
                          dest='l_degree',
                          type=int,
                          help=('Maximum legendre degree to add to the regression '
                                'matrix as nuisance. Default is 0, to account for '
                                'the degree of freedom lost in computing the SPC.'),
                          default=0)
    opt_regr.add_argument('-dmat', '--denoise-matrix',
                          dest='denoise_matrix',
                          nargs='*',
                          type=str,
                          help=('Complete path (absolute or relative) and filename '
                                'of denoising matrices to add to the regression model. '
                                'This option can be specified multiple times to '
                                'add multiple denoising matrices, but multiple '
                                'denoising matrices can be specified one after '
                                'the other, separated by a space.'),
                          default=None)
    opt_regr.add_argument('-scale', '--scale-factor',
                          dest='scale_factor',
                          type=float,
                          help=('Scale factor by which the beta maps will be divided '
                                'to create the CVR map output. Since BIDS '
                                'currently does not support mmHg as unit, if '
                                'using CO2 traces check their unit of measure '
                                'and their scaling factor to transform Volts into '
                                'mmHg. Use this option for other standardisations too.'),
                          default=None)

    opt_lreg = parser.add_argument_group('Optional Arguments for the lagged regression')
    opt_lreg.add_argument('-lm', '--lag-max',
                          dest='lag_max',
                          type=float,
                          help=('Maximum lag to consider during lag regression '
                                'in seconds. The same lag will be considered in '
                                'both directions.\n'
                                'Despite the code being python, the upper limit '
                                'is included in the computation. E.g. -lm 9 '
                                '-ls .3 means [-9, +9] (61 regressors).'),
                          default=None)
    opt_lreg.add_argument('-ls', '--lag-step',
                          dest='lag_step',
                          type=float,
                          help=('Lag step to consider during lagged regression '
                                'in seconds. Default is 0.3 seconds.'),
                          default=None)
    opt_lreg.add_argument('--legacy',
                          dest='legacy',
                          action='store_true',
                          help=('Use pythonic ranges, i.e. the upper limit '
                                'is excluded from the computation.\nE.g. -lm 9 '
                                '-ls .3 means [-9, +8.7] or [-9, +9) (60 regressors).'),
                          default=False)

    opt_regr = parser.add_argument_group('Optional Arguments to re-run a lagged '
                                         'regression (also useful to use a lag estimation '
                                         'on a different functional timeseries)')
    opt_regr.add_argument('-lmap', '--lag-map',
                          dest='lag_map',
                          type=str,
                          help=('Complete path (absolute or relative) and name '
                                'of a previously computed lag map to use in '
                                'lagged regression.'),
                          default=None)
    opt_regr.add_argument('-rdir', '--regr-dir',
                          dest='regr_dir',
                          type=str,
                          help=('Complete path (absolute or relative) and name '
                                'of previously computed lagged regressors to '
                                'use in a new lagged regression.'),
                          default=None)

    title_opt_conf = parser.add_argument_group('Optional Arguments to set up specific '
                                               'workflows')

    opt_conf = title_opt_conf.add_mutually_exclusive_group()
    opt_conf.add_argument('--brightspin',
                          dest='workflow_config',
                          action='store_const',
                          const='brightspin',
                          help=('Estimate CVR using a specific set of L-GLM parameters, '
                                'as used in:\n'
                                'S. Moia, et al., \'ICA-based denoising strategies in '
                                'breath-hold induced cerebrovascular reactivity mapping '
                                'with multi echo BOLD fMRI\' (2021), NeuroImage.\n'
                                'Same as setting --lag-max 9 --lag-step 0.3 '
                                '--legacy --r2full'),
                          default=None)
    opt_conf.add_argument('--brightspin-clinical',
                          dest='workflow_config',
                          action='store_const',
                          const='brightspin-clinical',
                          help=('Like "brightspin", but use a larger lag range.\n'
                                'Same as setting --lag-max 20 --lag-step 0.3 --r2full'),
                          default=None)
    opt_conf.add_argument('--baltimore',
                          dest='workflow_config',
                          action='store_const',
                          const='baltimore',
                          help=('Estimate CVR using the average timeseries in the '
                                '0.02-0.04 frequency spectrum, as used in:\n'
                                'P. Liu, et al., \'Cerebrovascular reactivity '
                                'mapping without gas challenges\' (2017), NeuroImage.\n'
                                'Same as setting --apply-filter -hf 0.04 -lf 0.02 '
                                '-skip_conv -skip_lagreg -co2 \'\' '),
                          default=None)
    opt_conf.add_argument('--baltimore-lag',
                          dest='workflow_config',
                          action='store_const',
                          const='baltimore-lag',
                          help=('Like "baltimore", but use a L-GLM instead.\n'
                                'Same as setting --apply-filter -hf 0.04 -lf 0.02 '
                                '-skip_conv -co2 \'\' '),
                          default=None)

    optional = parser.add_argument_group('Other Optional Arguments')
    optional.add_argument('-debug', '--debug',
                          dest='debug',
                          action='store_true',
                          help='Only print debugging info to log file. Default is False.',
                          default=False)
    optional.add_argument('-quiet', '--quiet',
                          dest='quiet',
                          action='store_true',
                          help='Only print warnings to log file. Default is False.',
                          default=False)
    optional.add_argument('-h', '--help', action='help',
                          help='Show this help message and exit')
    optional.add_argument('-v', '--version', action='version',
                          version=('%(prog)s ' + __version__))
    return parser


def _check_opt_conf(parser):
    """
    Check for particular configuration flags.
    
    Parameters
    ----------
    parser : argparse.ArgumentParser
        A parser with a 'workflow_config' and 'r2model' item inside.
    
    Returns
    -------
    parser : argparse.ArgumentParser
        If parser.workflow_config is None, returns the unmodified
        input parameter. Otherwise, set its items based on the flag.
        If parser.r2model is None, set it to 'full'
    
    Raises
    ------
    NotImplementedError
        If parser.workflow_config is not equal to a supported string.
        Which shouldn't happen, because this function should not be 
        called on its own.
    """
    if parser.workflow_config is not None:
        if parser.workflow_config == 'brightspin':
            parser.lag_max = 9
            parser.lag_step = 0.3
            parser.lagged_regression = True
            parser.run_conv = True
            parser.apply_filter = False
            parser.legacy = True
            parser.r2model = 'full'
        elif parser.workflow_config == 'brightspin-clinical':
            parser.lag_max = 20
            parser.lag_step = 0.3
            parser.lagged_regression = True
            parser.run_conv = True
            parser.apply_filter = False
            parser.r2model = 'full'
        elif parser.workflow_config == 'baltimore':
            parser.run_conv = False
            parser.apply_filter = True
            parser.lowcut = 0.02
            parser.highcut = 0.04
            parser.fname_co2 = None
            parser.lagged_regression = False
        elif parser.workflow_config == 'baltimore-lag':
            parser.run_conv = False
            parser.apply_filter = True
            parser.lowcut = 0.02
            parser.highcut = 0.04
            parser.fname_co2 = None
            parser.lagged_regression = True
        else:
            raise NotImplementedError(f'{parser.workflow_config} is not configured. '
                                      'In fact, you shouldn\'t see this message at all.')

    if parser.r2model is None:
        parser.r2model = 'full'

    del parser.workflow_config
    return parser


if __name__ == '__main__':
    raise RuntimeError('phys2cvr/cli/run.py should not be run directly;\n'
                       'Please `pip install` phys2cvr and use the '
                       '`phys2cvr` command')


"""
Copyright 2021, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""
Statistical module for phys2cvr.

Attributes
----------
LGR :
    Logger
"""

import logging
import os

import numpy as np
import matplotlib.pyplot as plt
import scipy.stats as sct

from phys2cvr import io
from phys2cvr.io import SET_DPI, FIGSIZE
from phys2cvr.signal import resample_signal


R2MODEL = ['full', 'partial', 'intercept', 'adj_full', 'adj_partial', 'adj_intercept']

LGR = logging.getLogger(__name__)
LGR.setLevel(logging.INFO)


def x_corr(func, co2, lastrep, firstrep=0, offset=0):
    """
    Cross correlation between `func` and `co2`.

    Parameters
    ----------
    func : np.ndarray
        Timeseries, must be SHORTER (or of equal length) than `co2`
    co2 : np.ndarray
        Second timeseries, can be LONGER than `func`
    lastrep : int
        Last index to take into account in `func`
    firstrep : int, optional
        First index totake into account in `func`
    offset : int, optional
        Optional amount of offset desired for `func`

    Returns
    -------
    float :
        Highest correlation
    int :
        Index of higher correlation
    xcorr : np.ndarray
        Full Xcorr

    Raises
    ------
    ValueError
        If `offset` is too high for the functional file
    """
    if len(func) + offset > len(co2):
        raise ValueError(f'The specified offset of {offset} is too high to '
                         f'compare func of length {len(func)} with co2 of '
                         f'length {len(co2)}')
    if firstrep + offset < 0:
        firstrep = -offset
    if lastrep + offset + len(func) > len(co2):
        lastrep = len(co2) - offset - len(func)

    xcorr = np.empty(lastrep+firstrep, dtype='float32')
    for i in range(firstrep, lastrep):
        xcorr[i] = np.corrcoef(func, co2[0+i+offset:len(func)+i+offset].T)[1, 0]

    return xcorr.max(), (xcorr.argmax() + firstrep + offset), xcorr


def get_regr(func_avg, petco2hrf, tr, freq, outname, lag_max=None,
             trial_len=None, n_trials=None, ext='.1D', lagged_regression=True,
             legacy=False):
    """
    Create regressor(s) of interest for nifti GLM.

    Parameters
    ----------
    func_avg : np.ndarray
        Functional timeseries (1D)
    petco2hrf : np.ndarray
        Regressor of interest
    tr : str, int, or float
        TR of timeseries
    freq : str, int, or float
        Sample frequency of petco2hrf
    outname : list or path
        Path to output directory for regressors.
    lag_max : int or float, optional
        Limits (both positive and negative) of the temporal area to explore,
        expressed in seconds.
        Default: 9 (i.e. ±9 seconds)
    trial_len : str or int, optional
        Length of each single trial for tasks that have more than one
        (E.g. BreathHold, CO2 challenges, ...)
        Used to improve cross correlation estimation.
        Default: None
    n_trials : str or int, optional
        Number of trials in the task.
        Default: None
    ext : str, optional
        Extension to be used for the exported regressors.
    lagged_regression : bool, optional
        Estimate regressors for each possible lag of `petco2hrf`.
        If True, the maximum number of regressors will be `(freq*lag_max*2)+1`
    legacy : bool, optional
        If True, exclude the upper lag limit from the regression estimation.
        If True, the maximum number of regressors will be `(freq*lag_max*2)`

    Returns
    -------
    petco2hrf_demean : np.ndarray
        The central, demeaned petco2hrf regressor.
    petco2hrf_shifts : np.ndarray
        The other shifted versions of the regressor.
    """
    # Setting up some variables
    first_tp, last_tp = 0, -1

    if trial_len and n_trials:
        # If both are specified, disregard two extreme _trial from matching.
        LGR.info(f'Specified {n_trials} trials lasting {trial_len} seconds')
        if n_trials > 2:
            LGR.info('Ignoring first trial to improve first bulk shift estimation')
            first_tp = int(trial_len*freq)
        else:
            LGR.info('Using all trials for bulk shift estimation')
        if n_trials > 3:
            LGR.info('Ignoring last trial to improve first bulk shift estimation')
            last_tp = first_tp*(n_trials-1)

    elif trial_len and not n_trials:
        LGR.warning('The length of trial was specified, but the number of '
                    'trials was not. Using all trials for bulk shift estimation')
    elif not trial_len and n_trials:
        LGR.warning('The number of trials was specified, but the length of '
                    'trial was not. Using all trials for bulk shift estimation')
    else:
        LGR.info('Using all trials for bulk shift estimation.')

    # Upsample functional signal
    func_upsampled = resample_signal(func_avg, 1/tr, freq)
    len_upd = func_upsampled.shape[-1]

    # Preparing breathhold and CO2 trace for Xcorr
    func_cut = func_upsampled[first_tp:last_tp]
    petco2hrf_cut = petco2hrf[first_tp:]

    nrep = petco2hrf_cut.shape[-1] - func_cut.shape[-1]

    _, optshift, xcorr = x_corr(func_cut, petco2hrf, nrep)
    LGR.info(f'First cross correlation estimated bulk shift at {optshift/freq} seconds')

    # Export estimated optimal shift in seconds
    with open(f'{outname}_optshift.1D', 'w') as f:
        print(f'{(optshift/freq):.4f}', file=f)

    petco2hrf_shift = petco2hrf[optshift:optshift+len_upd]

    # preparing for and exporting figures of shift
    time_axis = np.arange(0, nrep/freq, 1/freq)

    if nrep < time_axis.shape[-1]:
        time_axis = time_axis[:nrep]
    elif nrep > time_axis.shape[-1]:
        time_axis = np.pad(time_axis, (0, int(nrep - time_axis.shape[-1])), 'linear_ramp')

    plt.figure(figsize=FIGSIZE, dpi=SET_DPI)
    plt.plot(time_axis, xcorr)
    plt.title('optshift')
    plt.savefig(f'{outname}_optshift.png', dpi=SET_DPI)
    plt.close()

    plt.figure(figsize=FIGSIZE, dpi=SET_DPI)
    plt.plot(sct.zscore(petco2hrf_shift), '-', sct.zscore(func_upsampled), '-')
    plt.title('GM and shift')
    plt.savefig(f'{outname}_petco2hrf.png', dpi=SET_DPI)
    plt.close()

    petco2hrf_demean = io.export_regressor(petco2hrf_shift, freq, tr, outname,
                                           'petco2hrf', ext)

    # Initialise the shifts first.
    petco2hrf_shifts = None
    if lagged_regression and lag_max:
        outprefix = os.path.join(os.path.split(outname)[0], 'regr', os.path.split(outname)[1])
        os.makedirs(os.path.join(os.path.split(outname)[0], 'regr'), exist_ok=True)

        # Set num of fine shifts: 9 seconds is a bit more than physiologically feasible
        negrep = int(lag_max * freq)
        if legacy:
            posrep = negrep
        else:
            posrep = negrep + 1
        petco2hrf_shifts = np.empty([func_avg.shape[-1], negrep+posrep], dtype='float32')

        # Padding regressor for shift, and padding optshift too
        if (optshift - negrep) < 0:
            lpad = negrep - optshift
        else:
            lpad = 0

        if (optshift + posrep + len_upd) > petco2hrf.shape[-1]:
            rpad = (optshift + posrep + len_upd) - petco2hrf.shape[-1]
        else:
            rpad = 0

        petco2hrf_padded = np.pad(petco2hrf, (int(lpad), int(rpad)), 'mean')

        for n, i in enumerate(range(-negrep, posrep)):
            petco2hrf_lagged = petco2hrf_padded[optshift+lpad-i:optshift+lpad-i+len_upd]
            petco2hrf_shifts[:, n] = io.export_regressor(petco2hrf_lagged, freq, tr,
                                                         outprefix,
                                                         f'{n:04g}', ext)

    elif lagged_regression and lag_max is None:
        LGR.warning('The generation of lagged regressors was requested, '
                    'but the maximum lag was not specified. Skipping '
                    'lagged regressor generation.')
    else:
        LGR.info('Skipping lag regressors generation.')

    return petco2hrf_demean, petco2hrf_shifts


def get_legendre(degree, length):
    """
    Producesthe Legendre polynomials of order `degree`.

    Parameters
    ----------
    degree : int
        Highest order desired.
    length : int
        Number of samples of the polynomials.

    Returns
    -------
    legendre : np.ndarray
        A `degree`*`length` array with all the polynomials up to order `degree`
    """
    def _bonnet(d, x):
        if(d == 0):
            return np.ones_like(x)
        elif(d == 1):
            return x
        else:
            return ((2*d-1)*x*_bonnet(d-1, x)-(d-1)*_bonnet(d-2, x))/d

    x = np.linspace(-1, 1, length)
    legendre = np.empty([length, degree+1], dtype='float32')
    for n in range(degree+1):
        legendre[:, n] = _bonnet(n, x)
    return legendre


def regression(data, mask, regr, mat_conf, r2model='full', debug=False, x1D='mat.1D'):
    """
    Estimate regression parameters.

    Parameters
    ----------
    data : np.ndarray
        Dependent variable of the model (i.e. Y), as a 4D volume.
    mask : np.ndarray
        A 3D mask to reduce the number of voxels to run the regression for.
    regr : np.ndarray
        Regressor of interest
    mat_conf : np.ndarray
        Confounding effects (regressors)
    r2model : 'full', 'partial', 'intercept', 'adj_full', 'adj_partial', 'adj_intercept', optional
        R^2 to report.
        Possible models are:
            - 'full' (default)
                Use every regressor in the model, i.e. compare versus baseline 0
            - 'partial'
                Consider only `regr` in the model, i.e. compare versus baseline 
                composed by all other regressors (`mat_conf`)
            - 'intercept'
                Use every regressor in the model, but the intercept, i.e. compare 
                versus baseline intercept (Legendre polynomial order 0, a.k.a. 
                average signal)
            - 'adj_*'
                Adjusted R^2 version of normal counterpart
        Under normal conditions, while the R^2 value will be different between options, 
        a lagged regression based on any R^2 model will give the same results.
        This WILL NOT be the case if orthogonalisations between `regr` and `mat_conf`
        are introduced: a lagged regression based on `partial` might hold different 
        results.
        Default: 'full'

    Returns
    -------
    bout : np.ndarray
        Beta map
    tout : np.ndarray
        T-stats map
    rout : np.ndarray
        R^2 map

    Raises
    ------
    Exception
        If `mat_conf` and `regr` do not have at least one common dimension.
    """
    # Obtain data in 2d
    Ymat = data[mask]
    # Check that regr has "two" dimensions
    if len(regr.shape) < 2:
        regr = regr[..., np.newaxis]
    if regr.shape[0] != mat_conf.shape[0]:
        regr = regr.T
        if regr.shape[0] != mat_conf.shape[0]:
            raise ValueError('The provided confounding matrix does not match '
                             'the dimensionality of the PetCO2hrf regressor!')
    # Stack mat and solve least square
    # Note: Xmat is not currently demeaned within this function, so inputs should
    # already be demeaned (or within this function, demean all columns except zero order polynomial)
    Xmat = np.hstack([mat_conf, regr])

    if debug:
        os.makedirs(os.path.dirname(x1D), exist_ok=True)
        np.savetxt(x1D, Xmat, fmt='%.6f')
    # Xmat = mat-mat.mean(axis=0)
    betas, RSS, _, _ = np.linalg.lstsq(Xmat, Ymat.T, rcond=None)

    # compute t-values of betas (estimates)
    # first compute number of degrees of freedom
    df = Xmat.shape[0] - Xmat.shape[1]

    # compute sigma:
    # RSS = sum{[mdata - (X * betas)]^2}
    # sigma = RSS / Degrees_of_Freedom
    # RSS = np.sum(np.power(Ymat.T - np.dot(Xmat, betas), 2), axis=0)
    sigma = (RSS / df)
    sigma = sigma[..., np.newaxis]

    # Copmute std of betas:
    # C = (mat^T * mat)_ii^(-1)
    # std(betas) = sqrt(sigma * C)
    C = np.diag(np.linalg.pinv(np.dot(Xmat.T, Xmat)))
    C = C[..., np.newaxis]
    std_betas = np.sqrt(np.outer(C, sigma))
    tstats = betas / std_betas

    # Compute R^2 coefficient of multiple determination!
    r2model_support = False
    for model in R2MODEL:
        if r2model == model:
            r2model_support = True
    if not r2model_support:
        raise ValueError(f'{r2model} R^2 not supported. Supported R^2 models are {R2MODEL}')

    r2msg = ''
    # R^2 = 1 - RSS/TSS, (TSS = total sum of square)
    if 'full' in r2model:
        # Baseline model is 0 - this is what we're using ATM
        TSS = np.sum(np.power(Ymat, 2), axis=1)
    elif 'intercept' in r2model:
        # Baseline model is intercept: TSS is variance*samples
        TSS = Ymat.var(axis=1) * Ymat.shape[1]
    elif 'poly' in r2model:
        # Baseline model is legendre polynomials - or others: TSS is RSS of partial matrix
        # Could improve efficiency by moving this fitting step outside the regression loop
        # polynomials = Xmat[:, 0:4]
        # TSS = np.linalg.lstsq(polynomials, Ymat.T, rcond=None)[1]
        raise NotImplementedError('\'poly\' R^2 not implemented yet')
        r2msg = 'polynomial'
    elif 'partial' in r2model:
        pass

    if 'partial' in r2model:
        # We could also compute PARTIAL R square of regr instead (or on top)
        # See for computation: https://sscc.nimh.nih.gov/sscc/gangc/tr.html
        tstats_square = np.power(tstats, 2)
        r_square = (tstats_square / (tstats_square + df))[-1, :]
    else:
        r_square = np.ones(Ymat.shape[0], dtype='float32') - (RSS / TSS)

    if r2msg == '':
        r2msg = r2model
    if 'adj_' in r2model:
        # We could compute ADJUSTED R^2 instead
        r_square = 1-((1-r_square)*(Xmat.shape[0]-1)/(Xmat.shape[0]-Xmat.shape[1]))
        r2msg = f'adjusted {r2msg}'

    LGR.info(f'Adopting {r2msg} baseline to compute R^2.')

    # Assign betas, Rsquare and tstats to new volume
    bout = mask * 1.0
    tout = mask * 1.0
    rout = mask * 1.0
    bout[mask] = betas[-1, :]
    tout[mask] = tstats[-1, :]
    rout[mask] = r_square
    return bout, tout, rout


"""
Copyright 2021, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""
I/O and related utils.

Attributes
----------
EXT_1D : list
    List of supported TXT/1D file extensions, in lower case.
EXT_NIFTI : list
    List of supported nifti file extensions, in lower case.
FIGSIZE : tuple
    Figure size
SET_DPI : int
    DPI of the figure
LGR :
    Logger
"""

import logging

import nibabel as nib
import numpy as np

from phys2cvr import signal


SET_DPI = 100
FIGSIZE = (18, 10)
EXT_1D = ['.txt', '.csv', '.tsv', '.1d', '.par', '.tsv.gz']
EXT_NIFTI = ['.nii', '.nii.gz']

LGR = logging.getLogger(__name__)
LGR.setLevel(logging.INFO)


def if_declared_force_type(var, dtype, varname='an input variable', silent=False):
    """
    Make sure `var` is of type `dtype`.

    Parameters
    ----------
    var : str, int, or float
        Variable to change type of
    dtype : str
        Type to change `var` to
    varname : str, optional
        The name of the variable
    silent : bool, optional
        If True, don't return any message

    Returns
    -------
    int, float, str, list, or var
        The given `var` in the given `dtype`, or `var` if '' or None

    Raises
    ------
    NotImplementedError
        If dtype is not 'int', 'float', 'str', or 'list'
    """
    if var:
        if dtype == 'int':
            tmpvar = int(var)
        elif dtype == 'float':
            tmpvar = float(var)
        elif dtype == 'str':
            tmpvar = str(var)
        elif dtype == 'list':
            if type(var) == list:
                tmpvar = var
            else:
                tmpvar = [var]
        else:
            raise NotImplementedError(f'Type {dtype} not supported')

        if not silent:
            if type(tmpvar) != type(var):
                if varname != 'an input variable':
                    varname = f'variable {varname}'

                LGR.warning(f'Changing type of {varname} from {type(var)} to {dtype}')

        return tmpvar

    else:
        return var


def check_ext(all_ext, fname, remove=False):
    """
    Check which extension a file has, and possibly remove it.

    Parameters
    ----------
    all_ext : list
        All possible extensions to check within
    fname : str
        The filename to check
    remove : bool, optional
        Remove the extention from fname if it has one

    Returns
    -------
    str :
        If "remove" is True, return (extensionless) fname
    has_ext : boolean
        True if the extension is found, false otherwise
    """
    has_ext = False
    all_ext = if_declared_force_type(all_ext, 'list', silent=True)
    for ext in all_ext:
        if fname.lower().endswith(ext):
            has_ext = True
            break

    if remove:
        if has_ext:
            return fname[:-len(ext)], has_ext  # case insensitive solution
        else:
            return fname, has_ext
    else:
        return has_ext


def check_nifti_dim(fname, data, dim=4):
    """
    Remove extra dimensions.

    Parameters
    ----------
    fname : str
        The name of the file representing `data`
    data : np.ndarray
        The data which dimensionality needs to be checked
    dim : int, optional
        The amount of dimensions expected/desired in the data.

    Returns
    -------
    np.ndarray
        If `len(data.shape)` = `dim`, returns data.
        If `len(data.shape)` > `dim`, returns a version of data without the
        dimensions above `dim`.

    Raises
    ------
    ValueError
        If `data` has less dimensions than `dim`
    """
    if len(data.shape) < dim:
        raise ValueError(f'{fname} does not seem to be a {dim}D file. '
                         f'Plase provide a {dim}D nifti file.')
    if len(data.shape) > dim:
        LGR.warning(f'{fname} has more than {dim} dimensions. Removing D > {dim}.')
        for ax in range(dim, len(data.shape)):
            data = np.delete(data, np.s_[1:], axis=ax)

    return np.squeeze(data)


def load_nifti_get_mask(fname, is_mask=False, mask_dim=3):
    """
    Load a nifti file and returns its data, its image, and a 3d mask.

    Parameters
    ----------
    fname : str
        The filename to read in
    is_mask : bool, optional
        If the file contains a mask.
        Default: False
    mask_dim : int
        The number of dimensions expected in the mask

    Returns
    -------
    data : np.ndarray
        Data from nifti file.
    mask : np.ndarray
        If `is_mask` is False, np.ndarray of one dimension less than data,
        in which any element that has at least a value different from zero
        in the last dimension of `data` is True.
        If `is_mask` is True, mask is a boolean representation of data.
    img : nib.img
        Image object from nibabel.

    """
    img = nib.load(fname)
    LGR.info(f'Loading {fname}')
    data = img.get_fdata()

    if is_mask:
        data = check_nifti_dim(fname, data, dim=mask_dim)
        mask = (data < 0) + (data > 0)
    else:
        data = check_nifti_dim(fname, data)
        mask = np.squeeze(np.any(data, axis=-1))

    return data, mask, img


def export_regressor(petco2hrf_shift, freq, tr, outname, suffix='petco2hrf', ext='.1D'):
    """
    Export generated regressors for fMRI analysis.

    Parameters
    ----------
    petco2hrf_shift : np.ndarray
        The regressors that needs to be exported, in its original sample
    freq : int
        The frequency of the petco2hrf_shift (hence physiological data)
    tr : float
        The tr of the fMRI timeseries at which to export.
    outname : str or path
        Prefix of the output file - can contain a path.
    suffix : str, optional
        The suffix of the output file.
    ext : str, optional
        The extension of the output file.

    Returns
    -------
    petco2hrf_demean : np.ndarray
        Interpolated version of `petco2hrf_shift` in the sampling of the fMRI data.
    """
    petco2hrf_shift = signal.resample_signal(petco2hrf_shift, freq, 1/tr)
    petco2hrf_demean = petco2hrf_shift - petco2hrf_shift.mean()
    np.savetxt(f'{outname}_{suffix}{ext}', petco2hrf_demean, fmt='%.6f')

    return petco2hrf_demean


def export_nifti(data, img, fname):
    """
    Export a nifti file.

    Parameters
    ----------
    data : np.ndarray
        Data to be exported
    img : nib.img
        Nibabel image object
    fname : str or path
        Name of the output file
    """
    if fname.endswith('.nii.gz'):
        fname = fname[:-7]

    out_img = nib.Nifti1Image(data, img.affine, img.header)
    out_img.to_filename(f'{fname}.nii.gz')


"""
Copyright 2021, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
